"""Additional fields needed to disambiguate desired ElasticSearch output"""
import jsl

class TextField(jsl.StringField):
    """String field that gets analysed as text in ElasticSearch

    By default, JSL string fields are marked as opaque tokens by the mapping
    generator. Using this subclass instead of a normal jsl.StringField changes
    the ElasticSearch mapping generated to an analysed field without altering
    the normal JSL jsonschema output.
    """



"""High level JSL rendering tests"""
import jsl
from jsl_elasticsearch import TextField, render_es_template

import unittest

class TestRendering(unittest.TestCase):

    def _check_rendering(self, document, title, role, expected):
        rendered = render_es_template(document, title, role)
        self.assertEqual(rendered, expected)

    def test_primitive_fields(self):
        self._check_rendering(PrimitiveFields,
                              "project-primitive_fields",
                              "v1-0-0",
                              expected_primitive_template)

    def test_nested_documents(self):
        self._check_rendering(NestedDocuments,
                              "project-nested_documents",
                              "v1-0-0",
                              expected_nested_template)

    def test_versioned_documents(self):
        title = "project-multiversion_document"
        self._check_rendering(MultiversionDocument,
                              "project-multiversion_document",
                              "v1-0-0",
                              expected_multiversion_v1_0_0_template)
        self._check_rendering(MultiversionDocument,
                              "project-multiversion_document",
                              "v2-0-0",
                              expected_multiversion_v2_0_0_template)

#####################################################
# Example input documents for rendering tests
#####################################################

# Support required primitive fields
class PrimitiveFields(jsl.document.Document):
    single_str = jsl.StringField()
    str_array = jsl.ArrayField(jsl.StringField())
    text_data = TextField()
    text_array = jsl.ArrayField(TextField())
    single_int = jsl.IntField()
    int_array = jsl.ArrayField(jsl.IntField())
    single_float = jsl.NumberField()
    float_array = jsl.ArrayField(jsl.NumberField())

# Support nested documents
class NestedDocuments(jsl.document.Document):
    single_doc = jsl.DocumentField(PrimitiveFields, as_ref=True)
    doc_array = jsl.ArrayField(jsl.DocumentField(PrimitiveFields, as_ref=True))
    single_dict = jsl.DictField(dict(str_key=jsl.StringField(), int_key=jsl.IntField()))
    dict_array = jsl.ArrayField(jsl.DictField(dict(str_key=jsl.StringField(), int_key=jsl.IntField())))

# Support generating version specific mappings
ROLE_v1_0_0 = "v1-0-0"
ROLE_v2_0_0 = "v2-0-0"

ROLE_TITLE = jsl.roles.Var({
    ROLE_v1_0_0: "MultiversionDocument v1-0-0",
    ROLE_v2_0_0: "MultiversionDocument v2-0-0",
})

class MultiversionDocument(jsl.document.Document):
    class Options(object):
        title = ROLE_TITLE

    with jsl.roles.Scope(ROLE_v1_0_0) as v1:
        v1.subfield = jsl.fields.NumberField()
    with jsl.roles.Scope(ROLE_v2_0_0) as v2:
        v2.subfield = jsl.fields.IntField()

#####################################################
# Expected results from rendering example documents
#####################################################
expected_primitive_template = {
    "template": "project-primitive_fields-template_v1_0_0-*",
    "settings": {},
    "aliases": {
        "project-primitive_fields": {}
    },
    "mappings": {
        "content": {
            "properties": {
                "@timestamp": {
                    "type": "date",
                    "format": "dateOptionalTime"
                },
                "single_str": {
                    "index": "not_analyzed",
                    "type": "string",
                },
                "str_array": {
                    "index": "not_analyzed",
                    "type": "string",
                },
                "text_data": {
                    "type": "string"
                },
                "text_array": {
                    "type": "string"
                },
                "single_float": {
                    "type": "float"
                },
                "float_array": {
                    "type": "float"
                },
                "single_int": {
                    "type": "integer"
                },
                "int_array": {
                    "type": "integer"
                },
            }
        }
    }
}

expected_nested_template = {
  "template": "project-nested_documents-template_v1_0_0-*",
  "settings": {},
  "aliases": {
    "project-nested_documents": {}
  },
  "mappings": {
    "content": {
      "properties": {
        "@timestamp": {
          "type": "date",
          "format": "dateOptionalTime"
        },
        "single_doc": {
          "type": "nested",
          "properties": {
            "single_str": {
              "type": "string",
              "index": "not_analyzed",
            },
            "str_array": {
              "type": "string",
              "index": "not_analyzed",
            },
            "text_data": {
              "type": "string"
            },
            "text_array": {
              "type": "string"
            },
            "single_int": {
              "type": "integer"
            },
            "int_array": {
              "type": "integer"
            },
            "single_float": {
              "type": "float"
            },
            "float_array": {
              "type": "float"
            }
          }
        },
        "doc_array": {
          "type": "nested",
          "properties": {
            "single_str": {
              "type": "string",
              "index": "not_analyzed",
            },
            "str_array": {
              "type": "string",
              "index": "not_analyzed",
            },
            "text_data": {
              "type": "string"
            },
            "text_array": {
              "type": "string"
            },
            "single_int": {
              "type": "integer"
            },
            "int_array": {
              "type": "integer"
            },
            "single_float": {
              "type": "float"
            },
            "float_array": {
              "type": "float"
            }
          }
        },
        "single_dict": {
          "type": "nested",
          "properties": {
            "int_key": {
              "type": "integer"
            },
            "str_key": {
              "type": "string",
              "index": "not_analyzed",
            }
          }
        },
        "dict_array": {
          "type": "nested",
          "properties": {
            "int_key": {
              "type": "integer"
            },
            "str_key": {
              "type": "string",
              "index": "not_analyzed",
            }
          }
        }
      }
    }
  }
}


expected_multiversion_v1_0_0_template = {
  "template": "project-multiversion_document-template_v1_0_0-*",
  "settings": {},
  "aliases": {
    "project-multiversion_document": {}
  },
  "mappings": {
    "content": {
      "properties": {
        "@timestamp": {
          "type": "date",
          "format": "dateOptionalTime"
        },
        "subfield": {
          "type": "float"
        }
      }
    }
  }
}

expected_multiversion_v2_0_0_template = {
  "template": "project-multiversion_document-template_v2_0_0-*",
  "settings": {},
  "aliases": {
    "project-multiversion_document": {}
  },
  "mappings": {
    "content": {
      "properties": {
        "@timestamp": {
          "type": "date",
          "format": "dateOptionalTime"
        },
        "subfield": {
          "type": "integer"
        }
      }
    }
  }
}




# Project metadata
__version__ = "0.1.dev0"

__title__ = "jsl_elasticsearch"
__description__ = "Generate ElasticSearch templates from JSL"
__uri__ = "https://github.com/ncoghlan/jsl_elasticsearch"
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Nick Coghlan"
__email__ = "ncoghlan@gmail.com"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2016 Nick Coghlan"

# Public API
__all__ = [
    "TextField",
    "render_es_template",
    "render_es_mapping",
    "render_es_property",
]

from ._fields import (
    TextField,
)
from ._render import (
    render_es_template,
    render_es_mapping,
    render_es_property
)



from collections import OrderedDict
from functools import singledispatch

import jsl
from ._fields import TextField

def render_es_template(document, title, role, doc_type="content"):
    """Render an ElasticSearch time series template for given JSL document

    Template name is generated from the given *title* and *role*
    Document variables are resolved using the given *role*
    *doc_type* specifies the ElasticSearch mapping name (default: "content")
    """
    template = "{}-template_{}-*".format(title, role.replace("-", "_"))
    settings = {}
    aliases = {
        title: {}
    }
    doc_properties = render_es_mapping(document, role, add_timestamp=True)
    mappings = {
        doc_type: {
            "properties": doc_properties
        }
    }
    return OrderedDict((
        ("template", template),
        ("settings", settings),
        ("aliases", aliases),
        ("mappings", mappings),
    ))

def render_es_mapping(document, role, *, add_timestamp=False):
    """Render an ElasticSearch document mapping for given JSL document

    Document variables are resolved using the given *role*.
    Setting *add_timestamp* will add a ``dateOptionalTime`` formatted
    ``"@timestamp"`` field
    """
    result = OrderedDict()
    if add_timestamp:
        result["@timestamp"] = OrderedDict((
            ("type", "date"),
            ("format", "dateOptionalTime"),
        ))
    for name, field in document.resolve_and_iter_fields(role):
        result[name] = render_es_property(field, role)
    return result

@singledispatch
def render_es_property(jsl_field, role):
    """Render an individual JSL field as an ElasticSearch property"""
    raise NotImplementedError("Unknown JSL Field type: {!r}".format(type(jsl_field)))

@render_es_property.register(jsl.DocumentField)
def _render_document_field(jsl_field, role):
    return OrderedDict((
        ("type", "nested"),
        ("properties", render_es_mapping(jsl_field.document_cls, role)),
    ))

@render_es_property.register(jsl.DictField)
def _render_dict_field(jsl_field, role):
    properties, properties_role = jsl_field.resolve_attr('properties', role)
    result = OrderedDict()
    for name, field in sorted(properties.items()):
        result[name] = render_es_property(field, role)
    return OrderedDict((
        ("type", "nested"),
        ("properties", result),
    ))

@render_es_property.register(jsl.ArrayField)
def _render_array_field(jsl_field, role):
    # All ElasticSearch fields are implicitly arrays
    # so just render the contained field type
    return render_es_property(jsl_field.items, role)

@render_es_property.register(jsl.StringField)
def _render_string_field(jsl_field, role):
    # Note: marking the field as "not_analyzed" also implicitly switches
    # the default value of "doc_types" to True, enabling efficient lookups
    # based on the whole field value (rather than the field contents, which
    # is the default for analyzed strings)
    return OrderedDict((
        ("type", "string"),
        ("index", "not_analyzed"),
    ))

@render_es_property.register(TextField)
def _render_text_field(jsl_field, role):
    return {
        "type": "string",
    }

@render_es_property.register(jsl.IntField)
def _render_int_field(jsl_field, role):
    return {
        "type": "integer",
    }

@render_es_property.register(jsl.NumberField)
def _render_int_field(jsl_field, role):
    return {
        "type": "float",
    }
